Scriptus programs are written in JavaScript and are wrapped in a function declaration before being executed - this is why `return` acts as expected. All the usual JavaScript objects such as `Date` and `String` are present and correct, as are top-level functions such as `eval()`. However, none of the methods associated with in-browser JavaScript such as `alert()` are present. 

The Scriptus API is very simple, and can be divided into sections for interaction, program control, networking, and logging.

#Interaction

##say()

```javascript
say("this is a message");

say("this is a message to Tim", {to:"tim"});
```

This method sends the message provided to the person specified.

The second argument, providing additional parameters, is optional. `tim` above is an example Twitter screen name.

If no person is present, the message will be "said" to nobody in particular.

The return value of this function is a "message ID", which is an opaque string. You can use this message ID to `listen()` for replies to the message.

##listen()

```javascript
var heard = listen();

var heardFromHarper = listen({to:"harper"});
```

This method listens for messages. It returns the first message it receives as a string. If the `to` option is missing, it returns the first person to reply, or for Twitter, the first mention sent to the Twitter user.

You can tell who replied by looking at the `from` attribute of the message. In the above example, if `heardFromHarper` was not null, then `heardFromHarper.from` would equal "harper".

This method can also be used to listen to replies to specific messages.

```javascript
var chorusCall = say("Can I get a witness?");

var chorusResponse = listen({messageId:chorusCall});

var responseFromDiana = listen({messageId:chorusCall, to:"Diana"});
```

In the above examples, a message is first 'said', and then two responses are listened for. In the first example, Scriptus will return the first response from anyone who replies. In the second, only Diana is being listened to.

The `messageId` parameter is meant to allow script authors to listen for multiple replies to the same message. Be aware that deciding when to stop listening in a loop is important - e.g. after X replies or X amount of time.

##The `timeout` parameter and using durations in Scriptus

By default listen() waits for *24 hours* before returning to the program with `null` if no message has been received in that time.

This can be changed using the `timeout` option, which can be one of four things:

 * a number, 
 * a date as a string,
 * a JavaScript Date object,
 * or a duration.
 
These options can be used wherever a time or duration is possible in Scriptus:

```javascript
//For Leo, a number of hours (4):

var heardFromLeo = listen({to:"leo", timeout:4}); 

//For Charles, a date: May day 2020!
//format: yyyy-MM-dd HH:mm

var heardFromCharles = listen({to:"charles", timeout:"2020-05-01 07:00"}); 

//Buck Rogers gets a Date object:

var twentyFifthCentury = new Date();
twentyFifthCentury.setYear(2419);

var heardFromBuck = listen({to:"buck", timeout:twentyFifthCentury});

//For Kahlil, a duration - 4 years and a day:

var heardFromKahlil = listen({to:"kahlil", timeout:"4y, 1d"});
```

In the above snippet, `y` stands for years and `d` stands for days. Any number of durations can be combined. The comma is optional, so `4y, 1d` is the same as `4y 1d`. Other letters represent other things:

<table>
    <tr>
        <th>Letter</th>
        <th>Time period</th>
    </tr>
    <tr><td>s</td><td>Second</td></tr>
    <tr><td>m</td><td>Minute (lowercase)</td></tr>
    <tr><td>h</td><td>Hour</td></tr>
    <tr><td>d</td><td>Day (lowercase)</td></tr>
    <tr><td>w</td><td>Week</td></tr>
    <tr><td>M</td><td>Month (uppercase)</td></tr>
    <tr><td>q</td><td>Quarter (3 months)</td></tr>
    <tr><td>y</td><td>Year</td></tr>
    <tr><td>D</td><td>Decade (uppercase)</td></tr>
    <tr><td>C</td><td>Century</td></tr>
</table>

##ask()

```javascript
var deepThoughts = ask("Wherefore?");

var probablyNotDeep = ask("Dude! What does mine say?", {to:"chester"});

var unlikely = ask("A short novel please", {to:"victor", timeout:"17y"});
```

This method sends a message to a person in the style of `say`, and then awaits responses in the style of `listen`. The result, as with `listen`, is returned as a string, or `null` if the method times out.

If no person is specified using `to`, then the first reply to this question (or tweet) will be returned.

#Program control

The program control methods are basically taken from UNIX and are as follows:

##fork()

```javascript
var pid = fork();

if(pid == 0) {
	//this is the child;
	return "WAAAAAAA";
}
```

The `fork()` method is used to split the current process into two separate processes. Each process then continues execution. One of the processes is the 'child' and one of the processes is the 'parent'.

The child process will have `0` returned as the process identifier (or 'pid'), and the parent process will have returned the pid of the child.

The pid is not a number like in UNIX, but a UUID, represented as a string.

Both processes will continue to execute as normal. The parent process, knowing the identity of the child, can use the following function:

##wait()

```javascript
var pid = fork();

if(pid == 0) {
	//this is the child;
	return "WAAAAAAA";
}

var fromChild;

var receiver = function(result) {fromChild = result;}

var waitedFor = wait(receiver);

//fromChild == "WAAAAAA";
//waitedFor == pid;
```

What's happened here is a bit more tricky. The `fork()` call was used to split the process in two as above. Then an empty variable was declared for the results of the child process. Then another variable `receiver` was declared as a *function* that takes the result of the child process.

The wait function itself returns the pid of the process it waited for. If you have forked multiple processes and want to wait for one specific process, then you can select the process using its pid:

```javascript
var waitedFor = wait(receiver, pid);
```

If no pid is given, then the last process forked is waited for.

If no process was waited for, either because a bad pid was provided or no pid was provided and the process has not yet forked, `-1` is returned.

##kill()
```javascript
var pid = fork();

if(pid == 0) {
	//child
	sleep("2w");
	say("This will never happen!");
	return;
}

//parent
kill(pid);
```

Stops the process specified by the `pid` argument. If the specified process is sleeping or waiting, the process will simply be deleted.

If the process is currently executing, a best-effort is made to stop it from going any further. (I'm not completely sure it's working perfectly.)

##exec()

```javascript
var program = ask("C:\>_");

var pid = fork();
if(pid == 0) {
	exec(program, "arg1 arg2 arg3");
}
```

As in UNIX, the `exec()` call replaces the current process with a new process that is specified and executed with the arguments passed to the call. This call does not return. (To do: unless there is an error, in which case an exception will be thrown).

##exit()
```javascript
exit("pursued by a bear");
```

Quits the current process with the supplied argument as the return value, or `null` if not specified. The value can be any primitive or object.

This basically does the same as `return` in the top-level function but it can be called from anywhere in the process.

##sleep()
```javascript
sleep("8h");
```

The process will lie dormant for the specified amount of time, which as elsewhere may be a number of hours, a date, or a duration.

In many places it may be better to use `listen` so that a script can be woken up by prodding it if necessary.

<!--##TODO pipe()

When multiple processes have been created via `fork()` you may want them to communicate with each other without terminating. This is what pipe() is for. In UNIX, the method returns two "ids" at index 0 and 1 of the provided array `arr`, representing the start and end of the pipe respectively. In Scriptus, the supplied object will have the properties "in" and "out" set an ID. The IDs can then be used in the methods say(out) and listen(in) above in the 'person' parameter.-->

#Networking

The idea behind these methods is to allow the programming environment to interact with other services on the Internet. Right now the methods are very simple but adding more stuff in future is on the cards.

##get()
```javascript
var response = get("http://www.google.com/robots.txt");
var sslResponse = get("https://encrypted.google.com/robots.txt");
```

This command gets the result of an HTTP GET. In conjunction with `eval()` it can be used to import other scripts into your script. This can be used to setup libraries, such as parts of [dojo](http://www.dojotoolkit.org) or [jQuery](http://www.jQuery.com) for example.

When choosing URLs to import, I recommend that you use URLs to sites that you trust and are highly unlikely to change - for example, a link to a specific revision in a source code repo such as GitHub, Google Code or SourceForge.

At present, only HTTP and HTTPS URLs are supported. If this project takes off, SVN and Git URLs would be good to have too.

The HTTP GET is executed with a timeout of 60 seconds.

<!--TODO add the option to send & receive headers 

##TODO post()-->

#Logging

##log()
```javascript
log("Captain's log, Startdate 45240.1 To cross the Neutral Zone, I will require a cloaked ship...");
log(new Date());
```

The `log` method accepts one argument and then writes that argument's string respresentation to the log file `scriptus_programs.txt` under `logs/`. The log message will automatically prepend a timestamp, so you don't need to include one yourself.

##A note about `eval()`

Although `eval()` works, the Scriptus API methods as listed above cannot be executed within them. For example, this code wouldn't work:

```javascript
eval("say('Where ARE my socks?')")
```

Whereas this code should:

```javascript
eval("function(){say('Where are MY socks?');}")();
```

The reason for this is [documented here](http://mxr.mozilla.org/mozilla/source/js/rhino/testsrc/org/mozilla/javascript/tests/ContinuationsApiTest.java#203).
